@@ -3,13 +3,13 @@ use futures::future::{ok, ready};
 use futures::stream::Stream;
 use rusoto_core::ByteStream;
 use rusoto_s3::ListObjectsV2Output;
-use snafu::futures::TryStreamExt;
+use snafu::futures::TryFutureExt;
 use std::future::Future;
 use std::pin::Pin;
 use std::task::{Context, Poll};
 use tokio::io;
 
-pub type ListObjectsV2Result = Result<ListObjectsV2Output, RusotoError<ListObjectsV2Error>>;
+pub type ListObjectsV2Result = Result<(RequestReport, ListObjectsV2Output), err::Error>;
 
 /// A stream that can list objects, and (using member functions) delete or copy listed files.
 pub struct ListObjects<C, S> {
@@ -35,6 +35,29 @@ where
             prefix: self.prefix,
         }
     }
+
+    /// Calls an async closure on all the individual objects of the list operation
+    pub async fn process<P, F>(self, f: P) -> Result<(), Error>
+    where
+        P: Fn(Object) -> F + Clone,
+        F: Future<Output = ()>,
+    {
+        let ListObjects {
+            stream, prefix: _, ..
+        } = self;
+        stream
+            .try_filter_map(|response| ok(response.1.contents))
+            .map_ok(|x| stream::iter(x).map(Ok))
+            .try_flatten()
+            .try_for_each_concurrent(None, move |object| {
+                let f = f.clone();
+                async move {
+                    f(object).await;
+                    Ok(())
+                }
+            })
+            .await
+    }
     /// Download all listed objects - returns a stream of the contents.
     /// Used as a basis for other `download_all_*` functions.
     pub fn download_all_stream<R>(
@@ -57,7 +80,7 @@ where
         )));
         let n_retries = config.algorithm.n_retries;
         stream
-            .try_filter_map(|response| ok(response.contents))
+            .try_filter_map(|response| ok(response.1.contents))
             .map_ok(|x| stream::iter(x).map(Ok))
             .try_flatten()
             .try_filter_map(|obj| {
@@ -65,7 +88,6 @@ where
                 let Object { key, size, .. } = obj;
                 ok(key.and_then(|key| size.map(|size| (key, size))))
             })
-            .context(err::ListObjectsV2)
             .and_then(move |(key, size)| {
                 let (s3, timeout) = (s3.clone(), timeout.clone());
                 let request = GetObjectRequest {
@@ -80,10 +102,11 @@ where
                             let (s3, request) = (s3.clone(), request.clone());
                             Ok((
                                 async move { s3.get_object(request).context(err::GetObject).await },
-                                size as u64,
+                                size as usize,
                             ))
                         }
                     },
+                    |_, size| size,
                     n_retries,
                     timeout,
                 )
@@ -120,12 +143,20 @@ where
     }
     */
     /// Delete all listed objects.
-    /// In the `progress` closure's argument, the `size` file refers to the number of files
-    /// (attempted) deleted.
-    pub fn delete_all<P, F>(self, progress: P) -> impl Future<Output = Result<(), Error>>
+    /// ## The `progress` closure
+    /// The second (bool) argument says whether it's about a List or Delete request:
+    ///  - false: ListObjectsV2 request
+    ///  - true: DeleteObjects request
+    pub fn delete_all<P1, P2, F1, F2>(
+        self,
+        list_progress: P1,
+        delete_progress: P2,
+    ) -> impl Future<Output = Result<(), Error>>
     where
-        P: Fn(RequestReport) -> F + Clone + Send + Sync + 'static,
-        F: Future<Output = ()> + Send + 'static,
+        P1: Fn(RequestReport) -> F1 + Clone + Send + Sync + 'static,
+        P2: Fn(RequestReport) -> F2 + Clone + Send + Sync + 'static,
+        F1: Future<Output = ()> + Send + 'static,
+        F2: Future<Output = ()> + Send + 'static,
     {
         // For each ListObjectsV2Output, send a request to delete all the listed objects
         let ListObjects {
@@ -141,14 +172,27 @@ where
         )));
         let n_retries = config.algorithm.n_retries;
         stream
-            .filter_map(|response| ready(response.map(|r| r.contents).transpose()))
-            .map_err(|e| e.into())
+            .filter_map(|result| {
+                ready(
+                    result
+                        .map(|(report, response)| response.contents.map(|r| (report, r)))
+                        .transpose(),
+                )
+            })
+            .and_then(move |(report, contents)| {
+                // Report progress in listing objects
+                let list_progress = list_progress.clone();
+                async move {
+                    list_progress(report).await;
+                    Ok(contents)
+                }
+            })
             .try_for_each_concurrent(None, move |contents| {
-                let (s3, bucket, timeout, progress) = (
+                let (s3, bucket, timeout, del_progress) = (
                     s3.clone(),
                     bucket.clone(),
                     timeout.clone(),
-                    progress.clone(),
+                    delete_progress.clone(),
                 );
                 let objects = contents
                     .iter()
@@ -159,6 +203,7 @@ where
                         })
                     })
                     .collect::<Vec<_>>();
+                let n_objects = objects.len();
                 async move {
                     let (report, _) = s3_request(
                         move || {
@@ -167,7 +212,6 @@ where
                             async move {
                                 let (s3, bucket, objects) =
                                     (s3.clone(), bucket.clone(), objects.clone());
-                                let n_objects = objects.len();
                                 Ok((
                                     async move {
                                         s3.delete_objects(DeleteObjectsRequest {
@@ -181,24 +225,25 @@ where
                                         .map_err(|e| e.into())
                                         .await
                                     },
-                                    n_objects as u64,
+                                    n_objects,
                                 ))
                             }
                         },
+                        |_, size| size,
                         n_retries,
                         timeout.clone(),
                     )
                     .await?;
                     timeout.lock().await.update(&report);
-                    progress(report).await;
+                    del_progress(report).await;
                     Ok(())
                 }
             })
     }
     /// Flatten into a stream of Objects.
-    pub fn flatten(self) -> impl TryStream<Ok = Object, Error = RusotoError<ListObjectsV2Error>> {
+    pub fn flatten(self) -> impl TryStream<Ok = Object, Error = Error> {
         self.stream
-            .try_filter_map(|response| ok(response.contents))
+            .try_filter_map(|response| ok(response.1.contents))
             .map_ok(|x| stream::iter(x).map(Ok))
             .try_flatten()
     }
@@ -230,7 +275,7 @@ where
         let n_retries = config.algorithm.n_retries;
         let dest_bucket = dest_bucket.unwrap_or_else(|| bucket.clone());
         stream
-            .try_filter_map(|response| ok(response.contents))
+            .try_filter_map(|response| ok(response.1.contents))
             .map_ok(|x| stream::iter(x).map(Ok))
             .try_flatten()
             .try_filter_map(|obj| {
@@ -238,7 +283,6 @@ where
                 let Object { key, size, .. } = obj;
                 ok(key.and_then(|key| size.map(|size| (key, size))))
             })
-            .context(err::ListObjectsV2)
             .and_then(move |(key, size)| {
                 let (s3, timeout) = (s3.clone(), timeout.clone());
                 let request = CopyObjectRequest {
@@ -252,9 +296,10 @@ where
                         let (s3, request) = (s3.clone(), request.clone());
                         async move {
                             let (s3, request) = (s3.clone(), request.clone());
-                            Ok((async move{s3.copy_object(request).context(err::CopyObject).await}, size as u64))
+                            Ok((async move{s3.copy_object(request).context(err::CopyObject).await}, size as usize))
                         }
                     },
+                    |_, size| size,
                     n_retries,
                     timeout,
                 )
@@ -295,7 +340,7 @@ where
         let src_bucket = self.bucket.clone();
         let timeout = Arc::new(Mutex::new(TimeoutState::new(
             self.config.algorithm.clone(),
-            self.config.put_requests.clone(),
+            self.config.delete_requests.clone(),
         )));
         let n_retries = self.config.algorithm.n_retries;
         let s3 = self.s3.clone();
@@ -318,10 +363,11 @@ where
                                         .context(err::DeleteObject)
                                         .await
                                 },
-                                0,
+                                1,
                             ))
                         }
                     },
+                    |_, _| 1,
                     n_retries,
                     timeout,
                 )
@@ -360,7 +406,7 @@ where
     }
 }
 
-impl<S: S3 + Clone + Send + Sync + 'static> S3Algo<S> {
+impl<S: S3 + Clone + Send + Sync + Unpin + 'static> S3Algo<S> {
     /// List all objects with a certain prefix
     pub fn list_prefix(
         &self,
@@ -389,8 +435,13 @@ impl<S: S3 + Clone + Send + Sync + 'static> S3Algo<S> {
         request_factory: F,
     ) -> ListObjects<S, impl Stream<Item = ListObjectsV2Result> + Sized + Send>
     where
-        F: Fn() -> ListObjectsV2Request + Send + Sync + Clone,
+        F: Fn() -> ListObjectsV2Request + Send + Sync + Clone + Unpin + 'static,
     {
+        let n_retries = self.config.algorithm.n_retries;
+        let timeout = Arc::new(Mutex::new(TimeoutState::new(
+            self.config.algorithm.clone(),
+            self.config.delete_requests.clone(),
+        )));
         let s3_1 = self.s3.clone();
         let stream = futures::stream::unfold(
             // Initial state = (next continuation token, first request)
@@ -399,18 +450,39 @@ impl<S: S3 + Clone + Send + Sync + 'static> S3Algo<S> {
             //    - the stream will yield ListObjectsV2Output
             //      and stop when there is nothing left to list
             move |(cont, first)| {
-                let (s3, request_factory) = (s3_1.clone(), request_factory.clone());
+                let (s3, request_factory, timeout) =
+                    (s3_1.clone(), request_factory.clone(), timeout.clone());
                 async move {
                     if let (&None, false) = (&cont, first) {
                         None
                     } else {
-                        let result = s3
-                            .list_objects_v2(ListObjectsV2Request {
-                                continuation_token: cont,
-                                ..request_factory()
-                            })
-                            .await;
-                        let next_cont = if let Ok(ref response) = result {
+                        let result = s3_request(
+                            move || {
+                                let (s3, request_factory, cont) =
+                                    (s3.clone(), request_factory.clone(), cont.clone());
+                                // clone
+                                async move {
+                                    let (s3, request_factory, cont) =
+                                        (s3.clone(), request_factory.clone(), cont.clone());
+                                    Ok((
+                                        async move {
+                                            s3.list_objects_v2(ListObjectsV2Request {
+                                                continuation_token: cont,
+                                                ..request_factory()
+                                            })
+                                            .context(err::ListObjectsV2)
+                                            .await
+                                        },
+                                        1000, // Number of objects: assume worst case
+                                    ))
+                                }
+                            },
+                            |output, _| output.contents.as_ref().map(Vec::len).unwrap_or(0),
+                            n_retries,
+                            timeout.clone(),
+                        )
+                        .await;
+                        let next_cont = if let Ok((_, ref response)) = result {
                             response.next_continuation_token.clone()
                         } else {
                             None
@@ -434,17 +506,20 @@ impl<S: S3 + Clone + Send + Sync + 'static> S3Algo<S> {
 mod test {
     use super::*;
     use crate::test::rand_string;
+    use std::sync::atomic::AtomicUsize;
+    use std::sync::atomic::Ordering;
     #[tokio::test]
-    async fn test_s3_delete_files() {
+    async fn test_s3_delete_files_progress() {
         // Minio does paging at 10'000 fles, so we need more than that.
         // It means this test will take a minutes or two.
         let s3 = testing_s3_client();
         let algo = S3Algo::new(s3);
         let dir = rand_string(14);
+        let dir2 = dir.clone();
         const N_FILES: usize = 11_000;
         let files = (0..N_FILES).map(move |i| ObjectSource::Data {
             data: vec![1, 2, 3],
-            key: format!("{}/{}.file", dir, i),
+            key: format!("{}/{}.file", dir2, i),
         });
         algo.upload_files(
             "test-bucket".into(),
@@ -459,15 +534,75 @@ mod test {
         .await
         .unwrap();
 
+        let listed_files = Arc::new(AtomicUsize::new(0));
+        let deleted_files = Arc::new(AtomicUsize::new(0));
+        let listed_files2 = listed_files.clone();
+        let deleted_files2 = deleted_files.clone();
+
+        // Do one listing only to check the exact file names
+        let present = Arc::new(Mutex::new(std::collections::HashSet::new()));
+        algo.list_prefix("test-bucket".into(), String::new())
+            .process(|object| async {
+                let name = object.key.unwrap_or_else(|| "NONE".to_string());
+                println!("OBJ {}", name);
+                present.lock().await.insert(name);
+            })
+            .await
+            .unwrap();
+        let mut present = present.lock().await;
+
+        // All files are present
+        for i in 0..N_FILES {
+            let file_name = &format!("{}/{}.file", dir, i);
+            assert!(present.remove(file_name));
+        }
+
+        // No unexpected filesnames.
+        // Because once, it listed 11_200 files instead of 11_000
+        if !present.is_empty() {
+            println!("Left-over object names: {:?}", present);
+            panic!("Not empty");
+        }
+
+        // Assert that number of files is N_FILES
+        let count = algo
+            .list_prefix("test-bucket".into(), dir.clone())
+            .flatten()
+            .try_fold(0usize, |acc, _| ok(acc + 1))
+            .await
+            .unwrap();
+        assert_eq!(count, N_FILES);
+
         // Delete all
         algo.list_prefix("test-bucket".into(), String::new())
-            .delete_all(|_| async {})
+            .delete_all(
+                move |list_rep| {
+                    let n = list_rep.size as usize;
+                    println!("Listed {} items", n);
+                    let listed_files = listed_files2.clone();
+                    async move {
+                        listed_files.fetch_add(n, Ordering::Relaxed);
+                    }
+                },
+                move |del_rep| {
+                    let n = del_rep.size as usize;
+                    println!("Deleted {} items", n);
+                    let deleted_files = deleted_files2.clone();
+                    async move {
+                        deleted_files.fetch_add(n, Ordering::Relaxed);
+                    }
+                },
+            )
             .await
             .unwrap();
 
-        // List
+        // Assert number of objects listed and deleted
+        assert_eq!(listed_files.load(Ordering::Relaxed), N_FILES);
+        assert_eq!(deleted_files.load(Ordering::Relaxed), N_FILES);
+
+        // Assert that number of files is 0
         let count = algo
-            .list_prefix("test-bucket".into(), String::new())
+            .list_prefix("test-bucket".into(), dir)
             .flatten()
             .try_fold(0usize, |acc, _| ok(acc + 1))
             .await
